// spxlQuadGradient Effect Plugin
//
// FFGL effect plugin for Resolume Avenue and Resolume Arena to render a rectangular gradient (bilinear interpolation).
// Parameters specify the colors of the four corners.

// 2019-08-12 ffgl@subpixels.com
// - Created, based on the AddSubbtract sample FFGL2.1 effect plugin

#include "SpxlQuadGradientRGBEffect.h"
#include <misc/xorstr.h>

using namespace ffgl;
using namespace ffglex;

namespace spxl
{
namespace freeframe
{
namespace effect
{
////////////////////////////////////////////////////////////////////////////////////////////////////
// Plugin information
//
// Note definitions in external files:
//
//	{project build folder}\PluginInfo.props (just this project)
//	+ User macros:
//		FfglMajVer			(int)
//		FfglMinVer			(int)
//		FfglVer				(string)
//		PlugMajVer			(int)
//		PlugMinVer			(int)
//
//	{project build folder parent}\FFGLCommonPluginProperties.props (shared betweeen projects)
//	+ User macros:
//		AuthorName
//		AuthorEmail
//		SupportLink
//	+ Preprocessor definitions:
//		FFGLMAJVER = $(FfglMajVer)
//		FFGLMINVER = $(FfglMinVer)
//		FFGLVER = $(FfglVer)
//		PLATFORM = "$(FfglVer)-$(PlatformTarget)"
//		PLUGMAJVER = $(PlugMajVer)
//		PLUGMINVER = $(PlugMinVer)
//		PLUGVERSION = "v$(PlugMajVer).$(PlugMinVer)"
//		AUTHORNAME = "$(AuthorName)"
//		AUTHOREMAIL = "$(AuthorEmail)"
//		SUPPORTLINK = "$(SupportLink)"
////////////////////////////////////////////////////////////////////////////////////////////////////

static ffgl::PluginInfo pluginInfo(
	PluginFactory< QuadGradientRGBEffect >,// Create method
#ifdef _DEBUG
	"?QGr",            // Plugin unique ID of maximum length 4.
	"?QuadGradientRGB",// Plugin name
#else
	"_QGr",            // Plugin unique ID of maximum length 4.
	"_QuadGradientRGB",// Plugin name
#endif
	FFGLMAJVER,           // API major version number
	FFGLMINVER,           // API minor version number
	PLUGMAJVER,           // Plugin major version number
	PLUGMINVER,           // Plugin minor version number
	PluginType::FF_EFFECT,// Plugin type

	"Generate a gradient between four corner colors using bilinear interpolation",                   // Plugin description
	PLUGVERSION " (" PLATFORM ") by " AUTHORNAME " (" AUTHOREMAIL "). For support visit " SUPPORTLINK// About (see note re PluginInfo.props above)
);

////////////////////////////////////////////////////////////////////////////////////////////////////
// Shader code
//
// Biliner interpolation as described at https://en.wikipedia.org/wiki/Bilinear_interpolation
// f(x, y) = f(0, 0)(1 - x)(1 - y) + f(1, 0)x(1 - y) + f(0, 1)(1 - x)y + f(1, 1)xy;
//
// In this forumula:
//     f(0, 0) is the bottom left corner color
//     f(1, 0) is the bottom right corner color
//     f(0, 1) is the top left corner color
//     f(1, 1) is the top right corner color
//
// Ror Resolume 7, vertex Y-coordinates are 1 to -1, texture V-coordinates are 1 to 0 (top to bottom)
//
// In this plugin:
//     TL is tone1
//     TR is tone2
//     BL is tone3
//     BR is tone4
//
// Notes:
// Using p for the texture position, q is (1-p.x, 1-p.y) = 1-p
//
// color = (BL * (1.0 - p.x) * (1.0 - p.y)) + (BR * p.x * (1.0 - p.y)) + (TL * (1.0 - p.x) * p.y) + (TR * p.x * p.y)
//       = (BL * q.x         * q.y        ) + (BR * p.x * q.y        ) + (TL * q.x         * p.y) + (TR * p.x * p.y)
//       = (BL * q.x * q.y)                 + (BR * q.y * p.x)         + (TL * q.x * p.y)         + (TR * p.x * p.y)
//
// Collect together the multipliers:
// m = vec4(q.x * q.y, q.y * p.x, q.x * p.y, p.x * p.y)
//   = vec4(q.x, q.y, q.x, p.x) * vec4(q.y, p.x, p.y, p.y)
//   = vec4(q.xyx, p.x) * vec4(q.y, p.xyy)
//
// color = (BL * m.x )                      + (BR * m.y)               + (TL * m.z)               + (tone4 * m.w)
////////////////////////////////////////////////////////////////////////////////////////////////////

// Vertex shader
// - Pass through position and UV unmodified
// Note: xorstr() is used to obfuscate the shader code in the resulting DLL.
static std::string customVertexShaderCode = xorstr( R"(#version 410 core
layout( location = 0 ) in vec4 vPosition;
layout( location = 1 ) in vec2 vUV;

out vec2 i_uv;

void main()
{
	gl_Position = vPosition;
	i_uv = vUV;
}
)" );

// Fragment shader
// - Incoming i_uv is in range 0 to 1; need to multiply by maxUV for texture sampler, but not for bilinear interpolation
// - Scale fragColor alpha by input texture's alpha
// - Premultiply r,b,g by alpha
// Note: xorstr() is used to obfuscate the shader code in the resulting DLL.
static std::string fragmentShaderCode = xorstr( R"(
uniform vec2 maxUV;
void main(void)
{
	vec2 q = 1.0 - i_uv;
	vec4 m = vec4(q.xyx, i_uv.x) * vec4(q.y, i_uv.xyy);
	fragColor = tone3 * m.x + tone4 * m.y + tone1 * m.z + tone2 * m.w;
	fragColor.a *= texture( inputTexture, i_uv * maxUV ).a;
	fragColor.rgb *= fragColor.a;
}
)" );

QuadGradientRGBEffect::QuadGradientRGBEffect()
{
	// Four sets of parameters to specify four colors

	AddParam( new ParamFloat( "* TL R", "tone1", 1.0f, ParamFloat::Grouping::VEC4 ) );
	AddParam( new ParamFloat( "TL G", "", 0.0f ) );
	AddParam( new ParamFloat( "TL B", "", 0.0f ) );
	AddParam( new ParamFloat( "TL A", "", 1.0f ) );

	AddParam( new ParamFloat( "* TR R", "tone2", 1.0f, ParamFloat::Grouping::VEC4 ) );
	AddParam( new ParamFloat( "TR G", "", 1.0f ) );
	AddParam( new ParamFloat( "TR B", "", 0.0f ) );
	AddParam( new ParamFloat( "TR A", "", 1.0f ) );

	AddParam( new ParamFloat( "* BL R", "tone3", 0.0f, ParamFloat::Grouping::VEC4 ) );
	AddParam( new ParamFloat( "BL G", "", 1.0f ) );
	AddParam( new ParamFloat( "BL B", "", 1.0f ) );
	AddParam( new ParamFloat( "BL A", "", 1.0f ) );

	AddParam( new ParamFloat( "* BR R", "tone4", 0.0f, ParamFloat::Grouping::VEC4 ) );
	AddParam( new ParamFloat( "BR G", "", 0.0f ) );
	AddParam( new ParamFloat( "BR B", "", 1.0f ) );
	AddParam( new ParamFloat( "BR A", "", 1.0f ) );

	this->vertexShaderCode = customVertexShaderCode;
	SetFragmentShader( fragmentShaderCode );
}

QuadGradientRGBEffect::~QuadGradientRGBEffect()
{
}

}// namespace effect
}// namespace freeframe
}// namespace spxl
